@@ -12,31 +12,37 @@ from builtins import range
 import re, sys, os, shutil, subprocess#, xml.etree.ElementTree
 import requests
 import urllib.request, urllib.parse, urllib.error #TODO: substitute urllib with requests library
-from .misc import option, _abort
-from .doconce import errwarn
+from .misc import option, errwarn, _abort
 from doconce import globals
+from pygments.lexers import get_all_lexers, get_lexer_by_name
 
 format = None   # latex, pdflatex, html, plain, etc
 
 # Identifiers in the text used to identify code and math blocks
-_CODE_BLOCK = '<<<!!CODE_BLOCK'
-_MATH_BLOCK = '<<<!!MATH_BLOCK'
+_CODE_BLOCK = globals._CODE_BLOCK
+_MATH_BLOCK = globals._MATH_BLOCK
 
 # Chapter regex
 chapter_pattern = r'^=========\s*[A-Za-z0-9].+?========='
 
+# Table pattern
+pattern_table = r'^\|[\-lrcX]+\|'
+
 emoji_url = 'https://raw.githubusercontent.com/hplgit/doconce/master/bundled/emoji/png/'
 
 # Functions for creating and reading comment tags
 def begin_end_comment_tags(tag):
     return '--- begin ' + tag + ' ---', '--- end ' + tag + ' ---'
 
+
 def comment_tag(tag, comment_pattern='# %s'):
     return comment_pattern % tag
 
+
 def begin_comment_tag(tag, comment_pattern='# %s'):
     return comment_pattern % (begin_end_comment_tags(tag)[0])
 
+
 def end_comment_tag(tag, comment_pattern='# %s'):
     return comment_pattern % (begin_end_comment_tags(tag)[1])
 
@@ -66,6 +72,7 @@ envir_delimiter_lines = {
 
 _counter_for_html_movie_player = 0
 
+
 def internet_access(timeout=1):
     """Return True if internet is on, else False."""
     import urllib.request, urllib.error, urllib.parse, socket
@@ -73,13 +80,14 @@ def internet_access(timeout=1):
         # Check google.com with numerical IP-address (which avoids
         # DNS loopup) and set timeout to 1 sec so this does not
         # take much time (google.com should respond quickly)
-       #response = urllib2.urlopen('http://8.8.8.8', timeout=timeout)
-       response = urllib.request.urlopen('http://vg.no', timeout=timeout)
+       #response = urllib2.urlopen('https://8.8.8.8', timeout=timeout)
+       response = urllib.request.urlopen('https://vg.no', timeout=timeout)
        return True
     except (urllib.error.URLError, socket.timeout) as err:
         pass
     return False
 
+
 def safe_join(lines, delimiter):
     try:
         filestr = delimiter.join(lines) + '\n' # will fail if ord(char) > 127
@@ -92,6 +100,7 @@ def safe_join(lines, delimiter):
             errwarn(e)
             _abort()
 
+
 def fix_backslashes(text):
     """
     If some DocOnce text is read from a doc string
@@ -125,6 +134,7 @@ def where():
     import os
     return os.path.dirname(__file__)
 
+
 def is_file_or_url(filename, msg='checking existence of', debug=True):
     """
     Return "file" if filename is a file, "url" if filename
@@ -204,6 +214,7 @@ def has_copyright(filestr):
             break
     return copyright_, symbol
 
+
 def get_copyfile_info(filestr=None, copyright_filename=None, format=None):
     """Return copyright tuple in .filename.copyright."""
     # Copyright info
@@ -225,7 +236,7 @@ def get_copyfile_info(filestr=None, copyright_filename=None, format=None):
     if cr_info.get('license', None) is not None:
         cr_text += '. ' + cr_info['license']
     if cr_info.get('cite doconce', False):
-        url = 'https://github.com/hplgit/doconce'
+        url = 'https://github.com/doconce/doconce'
         if cr_text:
             cr_text += '. '
         cr_text += 'Made with '
@@ -240,6 +251,7 @@ def get_copyfile_info(filestr=None, copyright_filename=None, format=None):
 
     return cr_text
 
+
 def fix_ref_section_chapter(filestr, format):
     # .... see section ref{my:sec} is replaced by
     # see the section "...section heading..."
@@ -285,6 +297,7 @@ def fix_ref_section_chapter(filestr, format):
     # when it comes to section/chapter references.
     return filestr
 
+
 def indent_lines(text, format, indentation=' '*8, trailing_newline=True):
     """
     Indent each line in the string text, provided in a format for
@@ -298,12 +311,9 @@ def indent_lines(text, format, indentation=' '*8, trailing_newline=True):
         # it seems that Epytext is not capable of doing raw strings here)
         if re.search(r'\\n', text):
             comment_out = True
-            text = """\
-
-            NOTE: A verbatim block has been removed because
-                  it causes problems for Epytext.
-
-"""
+            text = ('\n'
+                    '            NOTE: A verbatim block has been removed because\n'
+                    '                  it causes problems for Epytext.\n\n')
             return text
 
     # indent X chars (choose X=6 for sufficient indent in lists)
@@ -312,6 +322,7 @@ def indent_lines(text, format, indentation=' '*8, trailing_newline=True):
         text += '\n'
     return text
 
+
 def unindent_lines(text, format=None, trailing_newline=True):
     """
     Unindent each line in the string text.
@@ -331,6 +342,7 @@ def unindent_lines(text, format=None, trailing_newline=True):
         text += '\n'
     return text
 
+
 def cite_with_multiple_args2multiple_cites(filestr):
     """Fix cite{key1,key2,key3} to cite{key1}cite[key2]cite[key3]."""
     cite_args = re.findall(r'cite\{(.+?)\}', filestr)
@@ -342,6 +354,7 @@ def cite_with_multiple_args2multiple_cites(filestr):
                 filestr = filestr.replace('cite{%s}' % arg, args)
     return filestr
 
+
 def table_analysis(table):
     """Return max width of each column."""
     # Find max no of columns
@@ -370,6 +383,7 @@ def table_analysis(table):
                 column_width[j] = max(column_width[j], len(column))
     return column_width
 
+
 def online_python_tutor(code, return_tp='iframe'):
     """
     Return URL (return_tp is 'url') or iframe HTML code
@@ -383,17 +397,18 @@ def online_python_tutor(code, return_tp='iframe'):
                                    'cumulative': 'false'})
         iframe = """\
 <iframe width="950" height="500" frameborder="0"
-        src="http://pythontutor.com/iframe-embed.html#code=%s&%s">
+        src="https://pythontutor.com/iframe-embed.html#code=%s&%s">
 </iframe>
 """ % (codestr, urlprm) # must treat code separately (urlencode adds chars)
         return iframe
     elif return_tp == 'url':
-        url = 'http://pythontutor.com/visualize.html#code=%s&mode=display&cumulative=false&heapPrimitives=false&drawParentPointers=false&textReferences=false&py=2&curInstr=0' % codestr
+        url = 'https://pythontutor.com/visualize.html#code=%s&mode=display&cumulative=false&heapPrimitives=false&drawParentPointers=false&textReferences=false&py=2&curInstr=0' % codestr
         url = url.replace('%', '\\%').replace('#', '\\#')
         return url
     else:
         raise ValueError('BUG!')
 
+
 def align2equations(filestr, format):
     """Turn align environments into separate equation environments."""
     if '{align' not in filestr:
@@ -485,6 +500,7 @@ def ref2equations(filestr):
     filestr = re.sub(r'XXX___(.+?)___XXX', r'\g<1>(ref{', filestr)
     return filestr
 
+
 def default_movie(m):
     """
     Replace a movie entry by a proper URL with text.
@@ -520,6 +536,7 @@ def default_movie(m):
        (caption, filename, moviehtml, moviehtml)
     return text
 
+
 def begin_end_consistency_checks(filestr, envirs):
     """Perform consistency checks: no of !bc must equal no of !ec, etc."""
     for envir in envirs:
@@ -556,10 +573,15 @@ def begin_end_consistency_checks(filestr, envirs):
 
 
 def remove_code_and_tex(filestr, format):
-    """
-    Remove verbatim and latex (math) code blocks from the file and
+    """Remove verbatim and latex (math) code blocks from the file and
     store separately in lists (code_blocks and tex_blocks).
-    The function insert_code_and_tex will insert these blocks again.
+    The functions insert_code_blocks and insert_tex_blocks will insert
+    these blocks again.
+
+    :param str filestr: text string
+    :param str format: output format
+    :return: filestr, code_blocks, code_block_types, tex_blocks
+    :rtype: str, list(str), list(str), list(str)
     """
     # Method:
     # store code and tex blocks in lists and substitute these blocks
@@ -571,14 +593,14 @@ def remove_code_and_tex(filestr, format):
     # ipynb (and future interactive executable documents) needs to
     # see if a code is to be executed or just displayed as text.
     # !bc *cod-t and !bc *pro-t is used to indicate pure text.
-    if format not in ('ipynb', 'matlabnb'):
-        filestr = re.sub(r'^!bc +([a-z0-9]+)-t', r'!bc \g<1>',
+    '''if format not in ('ipynb', 'matlabnb'):
+        filestr = re.sub(r'^!bc +([a-z0-9]+)', r'!bc \g<1>',  #-t
                          filestr, flags=re.MULTILINE)
     # !bc pypro-h for show/hide button
     if format not in ('html', 'sphinx'):
         filestr = re.sub(r'^!bc +([a-z0-9]+)-h', r'!bc \g<1>',
                          filestr, flags=re.MULTILINE)
-
+    '''
     # (recall that !bc can be followed by extra information that we must keep:)
     code = re.compile(r'^!bc(.*?)\n(.*?)^!ec *\n', re.DOTALL|re.MULTILINE)
 
@@ -689,6 +711,28 @@ def remove_code_and_tex(filestr, format):
 
     return filestr, code_blocks, code_block_types, tex_blocks
 
+
+def process_code_envir_postfix(environment):
+    """Extract any code envir postfix to code environments
+        ('hid','-h','-e','-t','out')
+
+    :param str environment: code environment e.g. 'pycod-e'
+    :return: postfix
+    :rtype: str
+    """
+    postfix = ""
+    if environment[-2:] in ['-h', '-e', '-t']:
+        # Show/Hide button (in html)
+        # Hide also in ipynb
+        # Code as text
+        postfix = environment[-2:]
+    elif environment[-3:] in ['hid', 'out']:
+        # Hide the cell
+        # Output cell
+        postfix = environment[-3:]
+    return postfix
+
+
 def add_labels_to_all_numbered_equations(tex_blocks):
     """
     Add a label with name _autoX, where X is an integer,
@@ -769,23 +813,35 @@ def add_labels_to_all_numbered_equations(tex_blocks):
     return tex_blocks
 
 
-def insert_code_and_tex(filestr, code_blocks, tex_blocks, format,
-                        complete_doc=True):
+def insert_code_blocks(filestr, code_blocks, format, complete_doc=True, remove_hid=True):
+    """ Insert code blocks in file
+
+    Insert in the DocOnce file verbatim code blocks previously extracted by
+    the `remove_code_and_tex` function and stored separately in lists.
+    :param str filestr: text string
+    :param list(str) code_blocks: list of code blocks
+    :param str format: output format
+    :param bool complete_doc: check with regex that the number of code blocks are
+    consistent, default True
+    :param bool remove_hid: remove hidden code blocks (e.g. !bc pyhid), default False
+    :return: filestr with code blocks in place
+    :rtype: str
+    """
     # Consistency check (only for complete documents):
-    # find no of distinct code and math blocks
+    # find no of distinct code blocks
     # (can be duplicates when solutions are copied at the end)
     pattern = r'^\d+ ' + _CODE_BLOCK
     code_lines = re.findall(pattern, filestr, flags=re.MULTILINE)
     n = len(set(code_lines))
     if complete_doc and len(code_blocks) != n:
         errwarn('*** error: found %d code block markers for %d initial code blocks' % (n, len(code_blocks)))
-        errwarn("""    Possible causes:
-           - mismatch of !bt and !et within one file, such that a !bt
-             swallows code
-           - mismatch of !bt and !et across files in multi-file documents
-           - !bc and !ec inside code blocks - replace by |bc and |ec
-    (run doconce on each individual file to locate the problem, then on
-     smaller and smaller parts of each file)""")
+        errwarn('    Possible causes:\n'
+                '           - mismatch of !bt and !et within one file, such that a !bt\n'
+                '             swallows code\n'
+                '           - mismatch of !bt and !et across files in multi-file documents\n'
+                '           - !bc and !ec inside code blocks - replace by |bc and |ec\n'
+                '    (run doconce on each individual file to locate the problem, then on\n'
+                '     smaller and smaller parts of each file)\n')
         numbers = list(range(len(code_blocks)))  # expected numbers in code blocks
         for e in code_lines:
             # remove number
@@ -798,24 +854,10 @@ def insert_code_and_tex(filestr, code_blocks, tex_blocks, format,
             errwarn('    Problem: did not find XX <<<!!CODE_BLOCK for XX in %s' % numbers)
 
         _abort()
-    pattern = r'^\d+ ' + _MATH_BLOCK
-    n = len(set(re.findall(pattern, filestr, flags=re.MULTILINE)))
-    if complete_doc and len(tex_blocks) != n:
-        errwarn('*** error: found %d tex block markers for %d initial tex blocks\nAbort!' % (n, len(tex_blocks)))
-        errwarn("""    Possible causes:
-           - mismatch of !bc and !ec within one file, such that a !bc
-             swallows tex blocks
-           - mismatch of !bc and !ec across files in multi-file documents
-           - !bt and !et inside code blocks - replace by |bt and |et
-    (run doconce on each file to locate the problem, then on
-     smaller and smaller parts of each file)""")
-        _abort()
-
-    from .misc import option
+    # Apply the max_bc_linelength option
     max_linelength = option('max_bc_linelength=', None)
     if max_linelength is not None:
         max_linelength = int(max_linelength)
-
         for i in range(len(code_blocks)):
             lines = code_blocks[i].splitlines()
             truncated = False
@@ -825,39 +867,69 @@ def insert_code_and_tex(filestr, code_blocks, tex_blocks, format,
                     truncated = True
             if truncated:
                 code_blocks[i] = '\n'.join(lines) + '\n'
-
-
+    # Loop through lines in input file
     lines = filestr.splitlines()
-
-    # Note: re.sub cannot be used because newlines, \nabla, etc
-    # are not handled correctly. Need str.replace.
-
     for i in range(len(lines)):
-        if _CODE_BLOCK in lines[i] or _MATH_BLOCK in lines[i]:
+        if _CODE_BLOCK in lines[i]:
             words = lines[i].split()
             # on a line: number block-indicator code-type
             n = int(words[0])
-            if _CODE_BLOCK in lines[i]:
-                words[1] = '!bc'
-                code = code_blocks[n]
-                lines[i] = ' '.join(words[1:]) + '\n' + code + '!ec'
-            if _MATH_BLOCK in lines[i]:
-                words[1] = '!bc'
-                math = tex_blocks[n]
-                lines[i] = '!bt\n' + math + '!et'
-
+            words[1] = '!bc'
+            code = code_blocks[n]
+            lines[i] = ' '.join(words[1:]) + '\n' + code + '!ec'
     filestr = safe_join(lines, '\n')
-
-    # All formats except sphinx and ipynb must remove !bc *hid blocks
-    # (maybe html will get the possibility to run hidden blocks)
-    if format not in ('sphinx', 'ipynb'):
+    # Remove hidden code blocks
+    if remove_hid:
         filestr = remove_hidden_code_blocks(filestr, format)
-
     return filestr
 
+
+def insert_tex_blocks(filestr, tex_blocks, format, complete_doc=True):
+    """ Insert code blocks in file
+
+    Insert in the DocOnce file verbatim and latex (math) code blocks
+    extracted by the `remove_code_and_tex` function and stored separately in lists.
+    :param str filestr: text string
+    :param list(str) tex_blocks: list of blocks
+    :param str format: output format
+    :param bool complete_doc: check with regex that the number of tex blocks
+    are consistent, default True
+    :return: filestr with tex blocks in place
+    :rtype: str
+    """
+    # Consistency check (only for complete documents):
+    # find no of distinct math blocks
+    # (can be duplicates when solutions are copied at the end)
+    pattern = r'^\d+ ' + _MATH_BLOCK
+    tex_lines = re.findall(pattern, filestr, flags=re.MULTILINE)
+    n = len(set(tex_lines))
+    if complete_doc and len(tex_blocks) != n:
+        errwarn('*** error: found %d tex block markers for %d initial tex blocks\nAbort!' % (n, len(tex_blocks)))
+        errwarn('    Possible causes:\n'
+                '           - mismatch of !bc and !ec within one file, such that a !bc\n'
+                '             swallows tex blocks\n'
+                '           - mismatch of !bc and !ec across files in multi-file documents\n'
+                '           - !bt and !et inside code blocks - replace by |bt and |et\n'
+                '    (run doconce on each file to locate the problem, then on\n'
+                '     smaller and smaller parts of each file)')
+        _abort()
+    # Loop through lines in input file
+    lines = filestr.splitlines()
+    for i in range(len(lines)):
+        if _MATH_BLOCK in lines[i]:
+            words = lines[i].split()
+            # on a line: number block-indicator code-type
+            n = int(words[0])
+            words[1] = '!bc'
+            math = tex_blocks[n]
+            lines[i] = '!bt\n' + math + '!et'
+    filestr = safe_join(lines, '\n')
+    return filestr
+
+
 def remove_hidden_code_blocks(filestr, format):
     """
-    Remove text encolsed in !bc *hid and !ec tags.
+    Remove text enclosed in !bc *hid and !ec tags.
     Some formats need this for executable code blocks to work,
     but they should be invisible in the text.
     """
@@ -865,6 +937,7 @@ def remove_hidden_code_blocks(filestr, format):
     filestr = re.sub(pattern, '', filestr, flags=re.MULTILINE|re.DOTALL)
     return filestr
 
+
 def doconce_exercise_output(
     exer,
     solution_header = '__Solution.__',
@@ -1111,6 +1184,8 @@ def doconce_exercise_output(
                         ['`%s`' % f for f in subex['file']]) + '.\n')
 
             if subex['ans_docend']:
+                # Enforce label to avoid any confusion with the sub-exercise
+                sol += '\nlabel{ansatend:' + exer['title'].lower().replace(' ', '_').replace(' ', '_') + '}\n'
                 # Similar to 'solution' above, but do not write solutions to main text
                 if exer['type'] != 'Example':
                     sol += '\n# ' + envir_delimiter_lines['ans_docend'][0] + '\n'
@@ -1121,6 +1196,8 @@ def doconce_exercise_output(
                     sol += '# ' + envir_delimiter_lines['ans_docend'][1] + '\n'
 
             if subex['sol_docend']:
+                # Enforce label to avoid any confusion with the sub-exercise
+                sol += '\nlabel{solatend:' + exer['title'].lower().replace(' ', '_').replace(' ', '_') + '}\n'
                 # Similar to 'solution' above, but do not write solutions to main text
                 if exer['type'] != 'Example':
                     sol += '\n# ' + envir_delimiter_lines['sol_docend'][0] + '\n'
@@ -1233,6 +1310,7 @@ def doconce_exercise_output(
         sol = ''
     return s, sol
 
+
 def plain_exercise(exer):
     return doconce_exercise_output(exer)
 
@@ -1271,24 +1349,24 @@ def bibliography(pubdata, citations, format='doconce'):
     text += '\n\n'
     return text
 
+
 def get_legal_pygments_lexers():
-    from pygments.lexers import get_all_lexers
     lexers = []
     for classname, names, dummy, dymmy in list(get_all_lexers()):
         for name in names:
             lexers.append(name)
     return lexers
 
+
 def has_custom_pygments_lexer(name):
-    from pygments.lexers import get_lexer_by_name
     if name == 'ipy':
         try:
-            get_lexer_by_name('ipy')
+            get_lexer_by_name(name)
         except Exception as e:
             errwarn('*** warning: !bc ipy used for IPython sessions, but')
             errwarn('    ipython is not supported for syntax highlighting!')
             errwarn('    install:')
-            errwarn('    git clone https://hplbit@bitbucket.org/hplbit/pygments-ipython-console.git; cd pygments-ipython-console; python setup.py install')
+            errwarn('    pip install pygments-ipython-console')
             errwarn(str(e))
             return False
     if name == 'doconce':
@@ -1298,9 +1376,7 @@ def has_custom_pygments_lexer(name):
             errwarn('*** warning: !bc do used for DocOnce code, but')
             errwarn('    not supported for syntax highlighting!')
             errwarn('    install:')
-            errwarn('    pip install --exists-action i -e git+https://github.com/hplgit/pygments-doconce#egg=pygments-doconce')
-            errwarn('\n    or manually:')
-            errwarn('    git clone https://github.com/hplgit/pygments-doconce.git; cd pygments-doconce; python setup.py install')
+            errwarn('    pip install pygments-doconce')
             errwarn(str(e))
             return False
     return True
@@ -1371,7 +1447,6 @@ def tikz2img(tikz_file, encoding='utf8', tikz_libs=None, pgfplots_libs=None):
 
 
     # TeX --> DVI
-    #print "TeX --> DVI"
     p = subprocess.Popen(['latex', '-output-directory='+tmp_dir,
                         '-interaction=nonstopmode',
                         tex_file],
@@ -1465,7 +1540,12 @@ TOC = {}
 ENVIRS = {}
 QUIZ = {}
 
-_linked_files = '''\s*"(?P<url>([^"]+?\.html?|[^"]+?\.html?\#[^"]+?|[^"]+?\.txt|[^"]+?\.tex|[^"]+?\.pdf|[^"]+?\.f|[^"]+?\.c|[^"]+?\.cpp|[^"]+?\.cxx|[^"]+?\.py|[^"]+?\.ipynb|[^"]+?\.java|[^"]+?\.pl|[^"]+?\.sh|[^"]+?\.csh|[^"]+?\.zsh|[^"]+?\.ksh|[^"]+?\.tar\.gz|[^"]+?\.tar|[^"]+?\.zip|[^"]+?\.f77|[^"]+?\.f90|[^"]+?\.f95|[^"]+?\.png|[^"]+?\.jpe?g|[^"]+?\.gif|[^"]+?\.pdf|[^"]+?\.flv|[^"]+?\.webm|[^"]+?\.ogg|[^"]+?\.mp4|[^"]+?\.mpe?g|[^"]+?\.e?ps|_static-?[^/]*/[^"]+?))"'''
+_linked_files = '\s*"(?P<url>([^"]+?\.html?|[^"]+?\.html?\#[^"]+?|[^"]+?\.txt|[^"]+?\.tex|[^"]+?\.pdf' \
+                '|[^"]+?\.f|[^"]+?\.c|[^"]+?\.cpp|[^"]+?\.cxx|[^"]+?\.py|[^"]+?\.ipynb|[^"]+?\.java|' \
+                '[^"]+?\.pl|[^"]+?\.sh|[^"]+?\.csh|[^"]+?\.zsh|[^"]+?\.ksh|[^"]+?\.tar\.gz|[^"]+?\.tar|' \
+                '[^"]+?\.zip|[^"]+?\.f77|[^"]+?\.f90|[^"]+?\.f95|[^"]+?\.png|[^"]+?\.jpe?g|[^"]+?\.gif|' \
+                '[^"]+?\.pdf|[^"]+?\.flv|[^"]+?\.webm|[^"]+?\.ogg|[^"]+?\.mp4|[^"]+?\.mpe?g|[^"]+?\.e?ps|' \
+                '_static-?[^/]*/[^"]+?))"'
 #_linked_files = '''\s*"(?P<url>([^"]+?))"'''  # any file is accepted
 
 abstract_names = '|'.join([globals.locale_dict[globals.locale_dict['language']][p]
@@ -1476,66 +1556,72 @@ INLINE_TAGS = {
     # first $ and space, comma, period, colon, semicolon, or question
     # mark after the enclosing $.
     'math':
-    r'%s\$(?P<subst>[^ `][^$`]*)\$%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
+        r'%s\$(?P<subst>[^ `][^$`]*)\$%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
 
     # $latex text$|$pure text alternative$
     'math2':
-    r'%s\$(?P<latexmath>[^ `][^$`]*)\$\|\$(?P<puretext>[^ `][^$`]*)\$%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
-    # simpler (not tested):
-    #r'%s\$(?P<latexmath>[^$]+?)\$\|\$(?P<puretext>[^$]+)\$%s' % \
-    #(globals.inline_tag_begin, globals.inline_tag_end),
+        r'%s\$(?P<latexmath>[^ `][^$`]*)\$\|\$(?P<puretext>[^ `][^$`]*)\$%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
+        # simpler (not tested):
+        # r'%s\$(?P<latexmath>[^$]+?)\$\|\$(?P<puretext>[^$]+)\$%s' % \
+        # (globals.inline_tag_begin, globals.inline_tag_end),
 
-    # *emphasized words*
+    # *emphasized words*  i.e. *italics*
     'emphasize':
-    r'%s\*(?P<subst>[^ `][^*`]*)\*%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
+        r'%s\*(?P<subst>[^ `][^*`]*)\*%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
 
-    # `verbatim inline text is enclosed in back quotes`
+    # `verbatim` inline text
     'verbatim':
-    r'%s`(?P<subst>[^ `][^`]*)`%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
-    #(globals.inline_tag_begin, r"(?P<end>($|[.,?!;:)}'\s|-]))"),
+        r'%s`(?P<subst>[^ `][^`]*)`%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
+        # (globals.inline_tag_begin, r"(?P<end>($|[.,?!;:)}'\s|-]))"),
 
-    # _underscore before and after signifies bold_
+    # _bold_
     'bold':
-    r'%s_(?P<subst>[^ `][^\]_`]*)_%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
+        r'%s_(?P<subst>[^ `][^\]_`]*)_%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
 
-    # color{col}{text} (\b is useful, but :.;`? is not word boundary)
+    # color{col}{text}
+    # (\b is useful, but :.;`? is not word boundary)
     'colortext':
-    r'\bcolor\{(?P<color>[^}]+?)\}\{(?P<text>[^}]+)\}',
+        r'\bcolor\{(?P<color>[^}]+?)\}\{(?P<text>[^}]+)\}',
 
-    # http://some.where.org/mypage<link text>  # old outdated syntax
+    # https://some.where.org/mypage<link text>  # old outdated syntax
     'linkURL':
-    r'%s(?P<url>https?://[^<\n]+)<(?P<link>[^>]+)>%s' % \
-    (globals.inline_tag_begin, globals.inline_tag_end),
+        r'%s(?P<url>https?://[^<\n]+)<(?P<link>[^>]+)>%s' % \
+        (globals.inline_tag_begin, globals.inline_tag_end),
 
-    'linkURL2':  # "some link": "https://bla-bla"
-    r'''"(?P<link>[^"]+?)" ?:\s*"(?P<url>(file:///|https?://|ftp://|mailto:).+?)"''',
-    #r'"(?P<link>[^>]+)" ?: ?"(?P<url>https?://[^<]+?)"'
+    # "some link": "https://bla-bla"
+    'linkURL2':
+        r'"(?P<link>[^"]+?)" ?:\s*"(?P<url>(file:///|https?://|ftp://|mailto:).+?)"',
+        # r'"(?P<link>[^>]+)" ?: ?"(?P<url>https?://[^<]+?)"'
 
-    'linkURL2v':  # verbatim link "`filelink`": "https://bla-bla"
-    r'''"`(?P<link>[^"]+?)`" ?:\s*"(?P<url>(file:///|https?://|ftp://|mailto:).+?)"''',
+    # verbatim link "`filelink`": "https://bla-bla"
+    'linkURL2v':
+        r'"`(?P<link>[^"]+?)`" ?:\s*"(?P<url>(file:///|https?://|ftp://|mailto:).+?)"',
 
-    'linkURL3':  # "some link": "some/local/file/name.html" or .txt/.pdf/.py/.c/.cpp/.cxx/.f/.java/.pl files
-    #r'''"(?P<link>[^"]+?)" ?:\s*"(?P<url>([^"]+?\.html?|[^"]+?\.txt|[^"]+?.pdf))"''',
-    r'''"(?P<link>[^"]+?)" ?:''' + _linked_files,
-    #r'"(?P<link>[^>]+)" ?: ?"(?P<url>https?://[^<]+?)"'
-    'linkURL3v':  # "`somefile`": "some/local/file/name.html" or .txt/.pdf/.py/.c/.cpp/.cxx/.f/.java/.pl files
-    r'''"`(?P<link>[^"]+?)`" ?:''' +  _linked_files,
+    # "some link": "some/local/file/name.html" or .txt/.pdf/.py/.c/.cpp/.cxx/.f/.java/.pl files
+    'linkURL3':
+        # r'"(?P<link>[^"]+?)" ?:\s*"(?P<url>([^"]+?\.html?|[^"]+?\.txt|[^"]+?.pdf))"',
+        r'"(?P<link>[^"]+?)" ?:' + _linked_files,
+
+    # "`somefile`": "some/local/file/name.html" or .txt/.pdf/.py/.c/.cpp/.cxx/.f/.java/.pl files
+    'linkURL3v':
+        r'"`(?P<link>[^"]+?)`" ?:' + _linked_files,
 
     'plainURL':
-    #r'"URL" ?: ?"(?P<url>.+?)"',
-    #r'"?(URL|url)"? ?: ?"(?P<url>.+?)"',
-    r'("URL"|"url"|URL|url) ?:\s*"(?P<url>.+?)"',
+        # r'"URL" ?: ?"(?P<url>.+?)"',
+        # r'"?(URL|url)"? ?: ?"(?P<url>.+?)"',
+        r'("URL"|"url"|URL|url) ?:\s*"(?P<url>.+?)"',
 
+    # [author name: comment]
     'inlinecomment':  # needs re.DOTALL
-    r'''\[(?P<name>[A-Za-z0-9 '+-]+?):(?P<space>\s+)(?P<comment>.*?)\]''',
-    # looks more robust for names with non-English characters,
-    # but caused problems (should not match \[ a:\quad ...\] and not footnotes)
-    #r'''(?<=[^\\])\[(?P<name>[^:\^]+?):(?P<space>\s+)(?P<comment>.*?)\]''',
+        r"\[(?P<name>[A-Za-z0-9 '+-]+?):(?P<space>\s+)(?P<comment>.*?)\]",
+        # looks more robust for names with non-English characters,
+        # but caused problems (should not match \[ a:\quad ...\] and not footnotes)
+        # r'(?<=[^\\])\[(?P<name>[^:\^]+?):(?P<space>\s+)(?P<comment>.*?)\]',
 
     # __Abstract.__ Any text up to a headline === or toc-like keywords
     # (TOC is already processed)
@@ -1543,69 +1629,90 @@ INLINE_TAGS = {
     # it before DATE (not recommended for papers)
     # 'abstract' is in doconce.py processed before chapter, section, etc
     'abstract':  # needs re.DOTALL | re.MULTILINE
-    r"""^\s*__(?P<type>%s).__\s*(?P<text>.+?)(?P<rest>TOC:|\\tableofcontents|Table of [Cc]ontents|DATE:|%% --- begin date|\\date\{|<!-- date|__[A-Z].+[.?:]__|^={3,9})""" % abstract_names,  # Abstract|Summary|Preface
+        (r"^\s*__(?P<type>%s).__\s*(?P<text>.+?)(?P<rest>TOC:|\\tableofcontents|"
+         r"Table of [Cc]ontents|DATE:|%% --- begin date|"
+         r"\\date\{|<!-- date|__[A-Z].+[.?:]__|^={3,9})") % abstract_names,  # Abstract|Summary|Preface
 
+    # keywords=integrals; SymPy
     'keywords':
-    r'^__Keywords.__\s+(?P<subst>.+)\s*$',
+        r'^__Keywords.__\s+(?P<subst>.+)\s*$',
 
-    # ======= Seven Equality Signs for Headline =======
     'section':
-    r'^={7}\s*(?P<subst>[^ =-].+?)\s*={7} *$',
+        r'^={7}\s*(?P<subst>[^ =-].+?)\s*={7} *$',
 
     'chapter':
-    r'^={9}\s*(?P<subst>[^ =-].+?)\s*={9} *$',
+        r'^={9}\s*(?P<subst>[^ =-].+?)\s*={9} *$',
 
     'subsection':
-    r'^={5}\s*(?P<subst>[^ =-].+?)\s*={5} *$',
+        r'^={5}\s*(?P<subst>[^ =-].+?)\s*={5} *$',
 
     'subsubsection':
-    # final \s is needed for latex to make \paragraph attached to text
-    # with no blank line
-    r'^={3}\s*(?P<subst>[^ =-].+?)\s*={3}\s*$',
+        # final \s is needed for latex to attach \paragraph to text without a blank line
+        r'^={3}\s*(?P<subst>[^ =-].+?)\s*={3}\s*$',
 
-    # __Two underscores for Inline Paragraph Title.__
+    # __Paragraph Title.__ and running text
     'paragraph':
-    r'(?P<begin>^)__(?P<subst>.+?)__(?P<space>(\n| +))',
+        r'(?P<begin>^)__(?P<subst>.+?)__(?P<space>(\n| *))',
+        #r'(?P<before>^[^_]*)__(?P<subst>.+?)__(?P<space>(\n| +))(?P<after>.*)$',
 
     # TITLE: My Document Title
     'title':
-    r'^TITLE:\s*(?P<subst>.+)\s*$',
+        r'^TITLE:\s*(?P<subst>.+)\s*$',
 
+    # AUTHOR: Some Name at Institution
     # AUTHOR: Some Name
     'author':
-    #r'^AUTHOR:\s*(?P<name>.+?)\s+at\s+(?P<institution>.+)\s*$',
-    # for backward compatibility (comma or at as separator):
-    r'^AUTHOR:\s*(?P<name>.+?)(,|\s+at\s+)(?P<institution>.+)\s*$',
+        # for backward compatibility (comma or at as separator):
+        r'^AUTHOR:\s*(?P<name>.+?)(,|\s+at\s+)?(?P<institution>.+)?\s*$',
 
     # DATE: Jan 27, 2010
     'date':
-    r'^DATE:\s*(?P<subst>.+)\s*$',
+        r'^DATE:\s*(?P<subst>.+)\s*$',
 
     # FIGURE:[filename, options] some caption text label{some:label}
     # (until blank line ^\s*$)
     'figure':
-    r'^FIGURE:\s*\[(?P<filename>[^,\]]+),?(?P<options>[^\]]*)\]\s*?(?P<caption>.*)$',
+        r'^FIGURE:\s*\[(?P<filename>[^,\]]+),?(?P<options>[^\]]*)\]\s*?(?P<caption>.*)$',
 
     # MOVIE:[filename, options] some caption text label{some:label}
     'movie':
-    r'^MOVIE:\s*\[(?P<filename>[^,\]]+),?(?P<options>[^\]]*)\]\s*?(?P<caption>.*)$',
-    'linebreak': '^(?P<text>.*)<linebreak> *$',
-    #'footnote':  # definition is in doconce.py since no regular re.sub in loop is to be performed
+        r'^MOVIE:\s*\[(?P<filename>[^,\]]+),?(?P<options>[^\]]*)\]\s*?(?P<caption>.*)$',
+
+    # text with<linebreak>
+    'linebreak':
+        '^(?P<text>.*)<linebreak> *$',
+
+    # footnotes[^footnote] ..  [^footnote]: my footnote
+    # 'footnote':  # definition is in doconce.py since no regular re.sub in loop is to be performed
     # The tilde is used in URLs and computer code
     # Must be substituted before inline math, color, etc., if the next
     # regex is to work (but then &nbsp;$math$ breaks later...)
-    #'non-breaking-space': r'(?<=[$A-Za-z0-9])~(?=[$A-Za-z0-9])',
+
+    # non-breaking space ~
     # This one allows HTML MathJax formulas and HTML tags to surround the ~
     # (i.e., after substitutions of $...$, color, etc.)
-    'non-breaking-space': r'(?<=[})>$A-Za-z0-9_`.])~(?=[{(\\<$A-Za-z0-9`:])',
-    'horizontal-rule': r'^----+$',
-    # ampersand1: Guns & Roses -> Guns {\&} Roses in latex
-    'ampersand1': r'(?P<pre>[A-Za-z0-9]) +& +(?P<post>[A-Za-z0-9])',  # \1 & \2
-    # Texas A & M (doconce) -> Texas A{\&}M in latex (no spaces around &)
-    'ampersand2': r' (?P<pre>[A-Z]) +& +(?P<post>[A-Z](?=\n|[ .,;-?:`]))',
-    'emoji': r'(\s):([a-z_]+):(\s)',
-    }
+    'non-breaking-space':
+        r'(?<=[})>$A-Za-z0-9_`.])~(?=[{(\\<$A-Za-z0-9`:])',
+
+    # -----
+    'horizontal-rule':
+        r'^----+$',
+
+    # ampersand: People & Robots -> People {\&} Robots
+    'ampersand1':
+        r'(?P<pre>[A-Za-z0-9]) +& +(?P<post>[A-Za-z0-9])',  # \1 & \2
+
+    # Texas A & M -> Texas A{\&}M in latex (no spaces around when single upper case letters on each side of &)
+    'ampersand2':
+        r' (?P<pre>[A-Z]) +& +(?P<post>[A-Z](?=\n|[ .,;-?:`]))',
+
+    'emoji':
+        r'(\s):([a-z_]+):(\s*)' if not option('no_emoji') else ''
+}
 
 INLINE_TAGS_SUBST = {}
 
 LIST_SYMBOL = {'*': 'itemize', 'o': 'enumerate', '-': 'description'}
+
+COMMANDS_TO_COMMENT = ['!split', '!bpop', '!epop', '!bslidecell', '!eslidecell',
+                '!bnotes', '!enotes']
\ No newline at end of file
